﻿using System;
using UnityEngine;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine.UI;
/***********************************************************************************************************************************
 * prestigitis HandPose 1.0.0
 * 
 * prestigitis_HandPose.cs 
 * add this script to a Person atom to individually adjust x, y, and z rotations for each bone in the hands. 
 * hand poses can be saved from either hand and restored to either hand.
 * 
 * This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License
 * https://creativecommons.org/licenses/by-nc-sa/4.0/
 *
 ***********************************************************************************************************************************/

namespace prestigitis {
    public class HandPose_1_0_0 : MVRScript {
/***********************************************************************************************************************************/
/* configuration settings 
/***********************************************************************************************************************************/
//UI options
private bool Verbose = false;                           //when loading a handpose file, send comments to message log
private bool StageOrientation = true;                   //use stage right/left orientation for UI
private float upperLimitDefault = 120f;                 //slider upper limit, default 120
private float lowerLimitDefault = -120f;                //slider lower limit, default -120
private bool ExpandOneFingerOnly = true;                //when true, closes other finger sliders when expanding one
Color buttonColorFingers = new Color(0.7f, 0.7f, 1f);   //color of finger toggle buttons
Color buttonColorReset = new Color(1f, 0.8f, 0.8f);     //color of hand pose reset button

//Save file options
private string HandPoseFileExtension = "handpose";                          //file extension for handpose files
private string HandPoseSubfolderName = "HandPose";                          //name of folder in [VaM]/Saves folder where handpose files are stored
string SaveFileHeaderString = "created by prestigitis_HandPose for VaM";   //header saved as comment in each handpose file
string SaveFileCommentCharacter = "#";                                      //comment character in handpose files
/***********************************************************************************************************************************/

        // IMPORTANT - DO NOT make custom enums. The dynamic C# complier crashes Unity when it encounters these for
        // some reason

        // IMPORTANT - DO NOT OVERRIDE Awake() as it is used internally by MVRScript - instead use Init() function which
        // is called right after creation

        private StringBuilder OutputBuilder = new StringBuilder();
        public Hand[] hand = new Hand[2]
        {
            new Hand(true), //right hand
            new Hand(false) //left hand
        };

        void UI_Update(Hand h)
        {
            bool UIColumnOrder = h.IsRight ^ StageOrientation;

            //remove all buttons for hand
            RemoveButton(h.SavePoseButton);
            RemoveButton(h.LoadPoseButton);
            RemoveButton(h.ResetButton);
            RemoveButton(h.UICarpalToggle);
            foreach (Segment s in h.GetSegments())
            {
                RemoveSlider(s.x);
                RemoveSlider(s.y);
                RemoveSlider(s.z);
            }
            foreach (Finger f in h.finger)
            {
                RemoveButton(f.UIToggleButton);
            }

            //add back all enabled buttons
            //generate save, load, and reset buttons
            h.SavePoseButton = CreateButton("Save " + h.name + " Hand Pose", UIColumnOrder);
            h.SavePoseButton.button.onClick.AddListener(() =>
            {
                SuperController.singleton.NormalizeMediaPath(SuperController.singleton.savesDir + "/" + HandPoseSubfolderName);
                SuperController.singleton.GetMediaPathDialog((string path) => {
                    SaveHandPose(h, path);
                }, HandPoseFileExtension, SuperController.singleton.savesDir + "/" + HandPoseSubfolderName, true, true, true);
                // Update the browser to be a Save browser
                uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
                browser.SetTextEntry(true);
                browser.fileEntryField.text = String.Format("{0}{1}.{2}", containingAtom.uid, ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString(), HandPoseFileExtension);
                browser.ActivateFileNameField();
            });

            h.LoadPoseButton = CreateButton("Load "+ h.name +" Hand Pose", UIColumnOrder);
            h.LoadPoseButton.button.onClick.AddListener(() => {
                SuperController.singleton.NormalizeMediaPath(SuperController.singleton.savesDir + "/" + HandPoseSubfolderName);
                SuperController.singleton.GetMediaPathDialog((string path) => {
                    LoadHandPose(h, path);
                }, HandPoseFileExtension, SuperController.singleton.savesDir + "/" + HandPoseSubfolderName, true, true, true);
            });

            h.ResetButton = CreateButton("Reset "+ h.name +" Hand", UIColumnOrder);
            h.ResetButton.buttonColor = buttonColorReset;
            h.ResetButton.button.onClick.AddListener(() => {
                h.ResetRotations();
            });

            //generate carpal toggle button
            h.UICarpalToggle = CreateButton(h.name + " Carpals", UIColumnOrder);
            h.UICarpalToggle.buttonColor = buttonColorFingers;
            h.UICarpalToggle.button.onClick.AddListener(() =>
            {
                bool b = !h.UIExpandedCarpals;
                h.ResetExpanded();
                h.UIExpandedCarpals = b;
                UI_Update(h);
            });
            if (h.UIExpandedCarpals)
            {
                foreach (Segment s in h.carpal)
                {
                    //generate x y and z sliders
                    CreateSlider(s.x, UIColumnOrder);
                    CreateSlider(s.y, UIColumnOrder);
                    CreateSlider(s.z, UIColumnOrder);
                }
            }
            foreach (Finger f in h.finger)
            {
                //generate finger toggle button
                f.UIToggleButton = CreateButton(h.name + " " + f.name, UIColumnOrder);
                f.UIToggleButton.buttonColor = buttonColorFingers;
                f.UIToggleButton.button.onClick.AddListener(() =>
                {
                    bool b = !f.UIExpanded;
                    h.ResetExpanded();
                    f.UIExpanded = b;
                    UI_Update(h);
                });
                if (f.UIExpanded)
                {
                    foreach (Segment s in f.segment)
                    {
                        //generate x y and z sliders
                        CreateSlider(s.x, UIColumnOrder);
                        CreateSlider(s.y, UIColumnOrder);
                        CreateSlider(s.z, UIColumnOrder);
                    }
                }
            }
        }

        public override void Init() {
            try {
                pluginLabelJSON.val = "prestigitis HandPose 1.0.0";
                hand[0].carpal[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal1").GetComponent<DAZBone>();
                hand[0].carpal[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal2").GetComponent<DAZBone>();
                hand[0].finger[0].segment[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rThumb1").GetComponent<DAZBone>();
                hand[0].finger[0].segment[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rThumb1/rThumb2").GetComponent<DAZBone>();
                hand[0].finger[0].segment[2].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rThumb1/rThumb2/rThumb3").GetComponent<DAZBone>();
                hand[0].finger[1].segment[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal1/rIndex1").GetComponent<DAZBone>();
                hand[0].finger[1].segment[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal1/rIndex1/rIndex2").GetComponent<DAZBone>();
                hand[0].finger[1].segment[2].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal1/rIndex1/rIndex2/rIndex3").GetComponent<DAZBone>();
                hand[0].finger[2].segment[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal1/rMid1").GetComponent<DAZBone>();
                hand[0].finger[2].segment[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal1/rMid1/rMid2").GetComponent<DAZBone>();
                hand[0].finger[2].segment[2].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal1/rMid1/rMid2/rMid3").GetComponent<DAZBone>();
                hand[0].finger[3].segment[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal2/rRing1").GetComponent<DAZBone>();
                hand[0].finger[3].segment[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal2/rRing1/rRing2").GetComponent<DAZBone>();
                hand[0].finger[3].segment[2].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal2/rRing1/rRing2/rRing3").GetComponent<DAZBone>();
                hand[0].finger[4].segment[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal2/rPinky1").GetComponent<DAZBone>();
                hand[0].finger[4].segment[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal2/rPinky1/rPinky2").GetComponent<DAZBone>();
                hand[0].finger[4].segment[2].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/rCollar/rShldr/rForeArm/rHand/rCarpal2/rPinky1/rPinky2/rPinky3").GetComponent<DAZBone>();

                hand[1].carpal[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal1").GetComponent<DAZBone>();
                hand[1].carpal[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal2").GetComponent<DAZBone>();
                hand[1].finger[0].segment[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lThumb1").GetComponent<DAZBone>();
                hand[1].finger[0].segment[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lThumb1/lThumb2").GetComponent<DAZBone>();
                hand[1].finger[0].segment[2].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lThumb1/lThumb2/lThumb3").GetComponent<DAZBone>();
                hand[1].finger[1].segment[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal1/lIndex1").GetComponent<DAZBone>();
                hand[1].finger[1].segment[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal1/lIndex1/lIndex2").GetComponent<DAZBone>();
                hand[1].finger[1].segment[2].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal1/lIndex1/lIndex2/lIndex3").GetComponent<DAZBone>();
                hand[1].finger[2].segment[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal1/lMid1").GetComponent<DAZBone>();
                hand[1].finger[2].segment[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal1/lMid1/lMid2").GetComponent<DAZBone>();
                hand[1].finger[2].segment[2].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal1/lMid1/lMid2/lMid3").GetComponent<DAZBone>();
                hand[1].finger[3].segment[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal2/lRing1").GetComponent<DAZBone>();
                hand[1].finger[3].segment[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal2/lRing1/lRing2").GetComponent<DAZBone>();
                hand[1].finger[3].segment[2].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal2/lRing1/lRing2/lRing3").GetComponent<DAZBone>();
                hand[1].finger[4].segment[0].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal2/lPinky1").GetComponent<DAZBone>();
                hand[1].finger[4].segment[1].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal2/lPinky1/lPinky2").GetComponent<DAZBone>();
                hand[1].finger[4].segment[2].bone = containingAtom.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/abdomen/abdomen2/chest/lCollar/lShldr/lForeArm/lHand/lCarpal2/lPinky1/lPinky2/lPinky3").GetComponent<DAZBone>();

                foreach (Hand h in hand)
                {

                    foreach (Segment s in h.carpal)
                    {
                        s.x = new JSONStorableFloat(h.name + " Carpal " + s.name + " X", 0f, s.UpdateRotation, lowerLimitDefault, upperLimitDefault, false);
                        RegisterFloat(s.x);
                        s.y = new JSONStorableFloat(h.name + " Carpal " + s.name + " Y", 0f, s.UpdateRotation, lowerLimitDefault, upperLimitDefault, false);
                        RegisterFloat(s.y);
                        s.z = new JSONStorableFloat(h.name + " Carpal " + s.name + " Z", 0f, s.UpdateRotation, lowerLimitDefault, upperLimitDefault, false);
                        RegisterFloat(s.z);
                    }
                    foreach (Finger f in h.finger)
                    {
                        foreach (Segment s in f.segment)
                        {
                            s.x = new JSONStorableFloat(h.name + " " + f.name + " " + s.name + " X", 0f, s.UpdateRotation, lowerLimitDefault, upperLimitDefault, false);
                            s.y = new JSONStorableFloat(h.name + " " + f.name + " " + s.name + " Y", 0f, s.UpdateRotation, lowerLimitDefault, upperLimitDefault, false);
                            s.z = new JSONStorableFloat(h.name + " " + f.name + " " + s.name + " Z", 0f, s.UpdateRotation, lowerLimitDefault, upperLimitDefault, false);
                            RegisterFloat(s.x);
                            RegisterFloat(s.y);
                            RegisterFloat(s.z);
                        }
                    }
                }

                foreach (Hand h in hand)
                {
                    h.UpdateSlidersFromBones();
                    UI_Update(h);
                }
            }
            catch (Exception e) {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        // OnDestroy is where you should put any cleanup
        // if you registered objects to supercontroller or atom, you should unregister them here
        void OnDestroy() {
            try
            {
                foreach (Hand h in hand)
                {
                    h.ResetRotations();
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        public void OnDisable()
        {
            foreach (Hand h in hand)
            {
                h.ResetBoneRotations();
                h.SetEnabled(false);
            }
        }
        public void OnEnable()
        {
            foreach (Hand h in hand)
            {
                h.SetEnabled(true);
                h.UpdateRotations();
            }
        }

        protected bool SaveFile(string path)
        {
            //save path must be in 'Saves' folder...
            SuperController.singleton.SaveStringIntoFile(path, OutputBuilder.ToString());
            return true;
        }
        
        protected void SaveHandPose(Hand h, string path)
        {
            try
            {
                if (path.Length == 0) return;
                OutputBuilder.Length = 0;
                OutputBuilder.AppendLine(SaveFileCommentCharacter + " " + SaveFileHeaderString);
                if (h.IsRight)
                {
                    foreach (Segment s in h.carpal)
                    {
                        OutputBuilder.AppendLine("Carpal " + s.name + ":" + s.x.val + " " + s.y.val + " " + s.z.val);
                    }
                    foreach (Finger f in h.finger)
                    {
                        foreach (Segment s in f.segment)
                        {
                            OutputBuilder.AppendLine(f.name + " " + s.name + ":" + s.x.val + " " + s.y.val + " " + s.z.val);
                        }
                    }
                }
                else
                {
                    foreach (Segment s in h.carpal)
                    {
                        OutputBuilder.AppendLine("Carpal " + s.name + ":" + s.x.val + " " + -s.y.val + " " + -s.z.val);
                    }
                    foreach (Finger f in h.finger)
                    {
                        foreach (Segment s in f.segment)
                        {
                            //invert Y and Z for left hand
                            OutputBuilder.AppendLine(f.name + " " + s.name + ":" + s.x.val + " " + -s.y.val + " " + -s.z.val);
                        }
                    }
                }
                if (!path.EndsWith("." + HandPoseFileExtension))
                {
                    path = path + "." + HandPoseFileExtension;
                }
                SaveFile(path);
                if (Verbose) SuperController.LogMessage("Saved HandPose File: " + path);
            }
            catch(Exception e)
            {
                SuperController.LogError("SaveHandPose Exception :" + e );
            }
        }
        protected void LoadHandPose(Hand h, string path)
        {
            try
            {
                if (path.Length == 0) return;
                //read file
                string LoadedHandPose = SuperController.singleton.ReadFileIntoString(path);
                string[] LoadedHandPoseLines = LoadedHandPose.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.None);
                foreach (string s in LoadedHandPoseLines)
                {
                    if (s.StartsWith(SaveFileCommentCharacter))
                    {
                        //its'a comment line, so ignore or send comment to logmessage
                        if (Verbose) SuperController.LogMessage("HandPose: " + s);
                    }
                    else if (s.Length == 0)
                    {
                        //ignore empty lines
                    }
                    else
                    {
                        //split string into segment id and values array
                        string[] segValues = s.Split(new string[] { ":" }, StringSplitOptions.None);
                        string[] segmentID = segValues[0].Split(new string[] { " " }, StringSplitOptions.None);
                        string[] values = segValues[1].Split(new string[] { " " }, StringSplitOptions.None);
                        //use linq to find the segment in hand h
                        var finger = h.finger.Where(f => f.name == segmentID[0]).FirstOrDefault();
                        Segment segment;
                        if (finger == null)
                        {
                            segment = h.carpal.Where(c => c.name == segmentID[1]).FirstOrDefault();
                        }
                        else
                        {
                            segment = finger.segment.Where(seg => seg.name == segmentID[1]).FirstOrDefault();
                        }
                        if (segment == null)
                        {
                            //couldn't find a carpal or finger segment
                        }
                        else if (h.IsRight)
                        {
                            //set segment bone values to values in values array
                            segment.x.val = float.Parse(values[0]);
                            segment.y.val = float.Parse(values[1]);
                            segment.z.val = float.Parse(values[2]);
                        }
                        else
                        {
                            //set segment bone values to values in values array
                            //for left hand, invert Y and Z values
                            segment.x.val = float.Parse(values[0]);
                            segment.y.val = -float.Parse(values[1]);
                            segment.z.val = -float.Parse(values[2]);
                        }
                    }
                    h.UpdateRotations();
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("LoadHandPose Exception " + e);
            }

        }
        private void StringTextbox(ref JSONStorableString output, string name, string start, JSONStorableString.SetStringCallback callback, bool rhs)
        {
            output = new JSONStorableString(name, start, callback);

            RegisterString(output);
            var textfield = CreateTextField(output, rhs);
            var input = textfield.gameObject.AddComponent<InputField>();
            input.textComponent = textfield.UItext;
            textfield.backgroundColor = Color.white;
            textfield.height = 0f;
            input.inputType = InputField.InputType.Standard;
            input.lineType = InputField.LineType.SingleLine;
            input.characterLimit = 31;
            input.textComponent.resizeTextForBestFit = false;
            output.inputField = input;
        }

        public class Segment
        {
            public bool enabled = true;
            public string name;
            public DAZBone bone;
            public JSONStorableFloat x; //ui slider for the segment's x rotation
            public JSONStorableFloat y; //ui slider for the segment's y rotation
            public JSONStorableFloat z; //ui slider for the segment's z rotation
            public Segment(string s)
            {
                name = s;
            }
            public void UpdateRotation(float f)
            {
                if (enabled) bone.baseJointRotation = new Vector3(x.val, y.val, z.val);
            }
            public void ResetRotation()
            {
                x.val = 0;
                y.val = 0;
                z.val = 0;
                UpdateRotation(0f);
            }
            public void ResetBoneRotation()
            {
                bone.baseJointRotation = Vector3.zero;
            }
        }
        public class Finger
        {
            public string name;
            public Segment[] segment = new Segment[3]
            {
                new Segment("1"),
                new Segment("2"),
                new Segment("3")
            };
            public UIDynamicButton UIToggleButton;
            public bool UIExpanded = false;
            public Finger(string s)
            {
                name = s;
            }
            public void UpdateRotations()
            {
                foreach (Segment s in segment) s.UpdateRotation(0f);
            }
            public void ResetRotations()
            {
                foreach (Segment s in segment) s.ResetRotation();
            }
            public void ResetBoneRotations()
            {
                foreach (Segment s in segment) s.ResetBoneRotation();
            }
        }
        public class Hand
        {
            public bool IsRight = true;
            public string name = "Right"; //right hand by default
            public Finger[] finger = new Finger[5]
            {
                new Finger("Thumb"),
                new Finger("Index"),
                new Finger("Middle"),
                new Finger("Ring"),
                new Finger("Pinky")
            };
            public Segment[] carpal = new Segment[2]
                {
                    new Segment("1"),
                    new Segment("2")
                };
            public UIDynamicButton UICarpalToggle;
            public bool UIExpandedCarpals = false;
            public UIDynamicButton SavePoseButton;
            public UIDynamicButton LoadPoseButton;
            public UIDynamicButton ResetButton;
            public JSONStorableString PoseNameField;
            public string PoseOutputFormat;

            public Hand(bool b)
            {
                IsRight = b;
                if (b) name = "Right";
                else name = "Left";
            }
            public void SetEnabled(bool b)
            {
                foreach (Finger f in finger)
                {
                    foreach (Segment s in f.segment) s.enabled = b;
                }
                foreach (Segment s in carpal) s.enabled = b;
            }
            public void UpdateRotations()
            {
                foreach (Finger f in finger) f.UpdateRotations();
                foreach (Segment s in carpal) s.UpdateRotation(0f);
            }
            public void ResetRotations()
            {
                foreach (Finger f in finger) f.ResetRotations();
                foreach (Segment s in carpal) s.ResetRotation();
            }
            public void ResetBoneRotations()
            {
                foreach (Finger f in finger) f.ResetBoneRotations();
                foreach (Segment s in carpal) s.ResetBoneRotation();
            }
            public void UpdateSlidersFromBones()
            {
                foreach (Segment s in GetSegments())
                {
                    s.x.val = s.bone.baseJointRotation.x;
                    s.y.val = s.bone.baseJointRotation.y;
                    s.z.val = s.bone.baseJointRotation.z;
                }
            }
            public List<Segment> GetSegments()
            {
                List<Segment> l = new List<Segment>();
                foreach (Finger f in finger)
                {
                    foreach (Segment s in f.segment) l.Add(s);
                }
                foreach (Segment s in carpal) l.Add(s);
                return l;
            }
            public void ResetExpanded()
            {
                foreach (Finger f in finger) f.UIExpanded = false;
                UIExpandedCarpals = false;
            }
        }
    }
}